#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : Python.
# @File         : DNN
# @Time         : 2020-03-13 13:42
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 
import tensorflow as tf
from tensorflow.python.keras import backend as K
from tensorflow.python.keras.initializers import Zeros, glorot_normal
from tensorflow.python.keras.layers import Layer
from tensorflow.python.keras.regularizers import l2


class DNN(Layer):

    def __init__(self, hidden_units, activation='relu', l2_reg=0, dropout_rate=0, use_bn=False, seed=1024, **kwargs):
        self.hidden_units = hidden_units
        self.activation = activation
        self.dropout_rate = dropout_rate
        self.seed = seed
        self.l2_reg = l2_reg
        self.use_bn = use_bn
        self.hidden_units_index = range(len(self.hidden_units))

        super(DNN, self).__init__(**kwargs)

    def build(self, input_shape):
        input_size = input_shape[-1]
        hidden_units = [int(input_size)] + list(self.hidden_units)

        # range(len(self.hidden_units))

        self.kernels = []
        self.bias = []

        for i in range(len(self.hidden_units)):
            kernel = self.add_weight(
                name='kernel' + str(i),
                shape=(hidden_units[i], hidden_units[i + 1]),
                initializer=glorot_normal(seed=self.seed),
                regularizer=l2(self.l2_reg),
                trainable=True
            )
            bias = self.add_weight(
                name='bias' + str(i),
                shape=(self.hidden_units[i],),
                initializer=Zeros(),
                trainable=True
            )

            self.kernels.append(kernel)
            self.bias.append(bias)

        if self.use_bn:
            self.bn_layers = [tf.keras.layers.BatchNormalization()] * len(self.hidden_units)

        self.dropout_layers = [tf.keras.layers.Dropout()] * len(self.hidden_units)

        self.activation_layers = [activation_layer(self.activation)] * len(self.hidden_units)

        super(DNN, self).build(input_shape)  # Be sure to call this somewhere! 传递信息？

    def call(self, inputs, training=None, **kwargs):

        deep_input = inputs
        for i in range(len(self.hidden_units)):
            fc = tf.nn.bias_add(tf.tensordot(deep_input, self.kernels[i], axes=(-1, 0)), self.bias[i])
            fc = self.bn_layers[i](fc, training=training) if self.use_bn else fc
            fc = self.activation_layers[i](fc)
            fc = self.dropout_layers[i](fc, training=training)
        return fc

    def compute_output_shape(self, input_shape):
        if len(self.hidden_units) > 0:
            shape = input_shape[:-1] + (self.hidden_units[-1],)
        else:
            shape = input_shape

        return tuple(shape)

    def get_config(self, ):
        config = {
            'activation': self.activation,
            'hidden_units': self.hidden_units,
            'l2_reg': self.l2_reg,
            'use_bn': self.use_bn,
            'dropout_rate': self.dropout_rate,
            'seed': self.seed
        }
        base_config = super(DNN, self).get_config()  # TODO: 更新配置信息
        return {**base_config, **config}
        # return dict(list(base_config.items()) + list(config.items()))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : timer
# @Time         : 2020-02-12 00:05
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :

# class MyResource:
#     def query(self):
#         print('query data')
#
# @contextmanager
# def make_myresource():
#     print('start to connect')
#     yield MyResource()
#     print('end connect')
#     pass
# with make_myresource() as r:
#      r.query()

import time
from contextlib import contextmanager
from ..utils import cprint


@contextmanager
def timer(task_name="timer"):
    print('\n')
    cprint(f"⏰ {task_name} started")
    t0 = time.time()
    yield
    cprint(f"⏰ {task_name} done in {time.time() - t0:.3f} seconds")


if __name__ == '__main__':
    with timer() as t:
        print('xxx')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : __init__.py
# @Time         : 2019-09-10 15:22
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 
from functools import partial
from collections import Iterable

from keras.layers import Dense


# last layer: 'softmax' or 'sigmoid'
def getDenseList(unitsList, activation=None):
    assert isinstance(unitsList, Iterable)
    dense = partial(Dense, activation=activation)
    return map(dense, unitsList)





#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : __init__.py
# @Time         : 2019-09-10 14:48
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  :
import numpy as np
from .cprint import cprint
from .timer import timer


def noramlize(x):
    return x / np.linalg.norm(x, 2, axis=len(x.shape) > 1, keepdims=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : DataLoader
# @Time         : 2019-10-21 13:01
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import tensorflow as tf

"""
https://www.cnblogs.com/gengyi/p/11107492.html
https://tensorflow.google.cn/guide/data?hl=zh_cn
https://cs230-stanford.github.io/tensorflow-input-data.html#shuffle-and-repeat

tf.data.Dataset.zip
"""


class DataGenerator(object):

    def __init__(self, data_type='train', batchsize=128):
        self.data_type = data_type
        self.batchsize = batchsize

    def from_numpy(self, tensors, buffer_size=10000, seed=666):
        dataset = tf.data.Dataset.from_tensor_slices(tensors).batch(self.batchsize)
        if self.data_type in ('train', 'valid'):
            dataset = dataset.shuffle(buffer_size, seed)

        return dataset



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : shake
# @Time         : 2019-09-14 23:27
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

import matplotlib.pyplot as plt
from tqdm import tqdm
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import train_test_split
import numba
import numpy as np


class Shake(object):

    def __init__(self, oof_preds, y_true, feval=roc_auc_score):
        self.oof_preds = oof_preds
        self.y_true = y_true
        self.feval = feval

    def plot_difference(self, n_splits=100, test_size=0.7):
        self.cal(n_splits, test_size)
        self.public_private = np.array(self.metric_public) - np.array(self.metric_private)

        plt.figure(figsize=(10, 6))
        plt.hist(self.public_private, bins=50)
        plt.title('(Public - Private) scores')
        plt.xlabel(f'{self.feval} score difference')
        plt.show()

    @numba.jit()
    def cal(self, n_splits=100, test_size=0.7):
        self.metric_public = []
        self.metric_private = []
        for rs in tqdm(range(n_splits)):
            _ = train_test_split(self.oof_preds, self.y_true, test_size=test_size, random_state=rs)
            y_preds_public, y_preds_private, y_public, y_private = _
            self.metric_public.append(self.feval(y_public, y_preds_public))
            self.metric_private.append(self.feval(y_private, y_preds_private))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : F1Optimizer
# @Time         : 2019-09-11 12:03
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


import numpy as np
import pandas as pd
import matplotlib.pylab as plt
from datetime import datetime

'''
This kernel implements the O(n²) F1-Score expectation maximization algorithm presented in
"Ye, N., Chai, K., Lee, W., and Chieu, H.  Optimizing F-measures: A Tale of Two Approaches. In ICML, 2012."

It solves argmax_(0 <= k <= n,[[None]]) E[F1(P,k,[[None]])]
with [[None]] being the indicator for predicting label "None"
given posteriors P = [p_1, p_2, ... , p_n], where p_1 > p_2 > ... > p_n
under label independence assumption by means of dynamic programming in O(n²).


c: https://gist.github.com/tkm2261/3f44886b07f87ca953fba3d328ca256b
'''


class F1Optimizer():
    def __init__(self):
        pass

    @staticmethod
    def get_expectations(P, pNone=None):
        expectations = []
        P = np.sort(P)[::-1]

        n = np.array(P).shape[0]
        DP_C = np.zeros((n + 2, n + 1))
        if pNone is None:
            pNone = (1.0 - P).prod()

        DP_C[0][0] = 1.0
        for j in range(1, n):
            DP_C[0][j] = (1.0 - P[j - 1]) * DP_C[0, j - 1]

        for i in range(1, n + 1):
            DP_C[i, i] = DP_C[i - 1, i - 1] * P[i - 1]
            for j in range(i + 1, n + 1):
                DP_C[i, j] = P[j - 1] * DP_C[i - 1, j - 1] + (1.0 - P[j - 1]) * DP_C[i, j - 1]

        DP_S = np.zeros((2 * n + 1,))
        DP_SNone = np.zeros((2 * n + 1,))
        for i in range(1, 2 * n + 1):
            DP_S[i] = 1. / (1. * i)
            DP_SNone[i] = 1. / (1. * i + 1)
        for k in range(n + 1)[::-1]:
            f1 = 0
            f1None = 0
            for k1 in range(n + 1):
                f1 += 2 * k1 * DP_C[k1][k] * DP_S[k + k1]
                f1None += 2 * k1 * DP_C[k1][k] * DP_SNone[k + k1]
            for i in range(1, 2 * k - 1):
                DP_S[i] = (1 - P[k - 1]) * DP_S[i] + P[k - 1] * DP_S[i + 1]
                DP_SNone[i] = (1 - P[k - 1]) * DP_SNone[i] + P[k - 1] * DP_SNone[i + 1]
            expectations.append([f1None + 2 * pNone / (2 + k), f1])

        return np.array(expectations[::-1]).T

    @staticmethod
    def maximize_expectation(P, pNone=None):
        expectations = F1Optimizer.get_expectations(P, pNone)

        ix_max = np.unravel_index(expectations.argmax(), expectations.shape)
        max_f1 = expectations[ix_max]

        predNone = True if ix_max[0] == 0 else False
        best_k = ix_max[1]

        return best_k, predNone, max_f1

    @staticmethod
    def _F1(tp, fp, fn):
        return 2 * tp / (2 * tp + fp + fn)

    @staticmethod
    def _Fbeta(tp, fp, fn, beta=1.0):
        beta_squared = beta ** 2
        return (1.0 + beta_squared) * tp / ((1.0 + beta_squared) * tp + fp + beta_squared * fn)


def print_best_prediction(P, pNone=None):
    print("Maximize F1-Expectation")
    print("=" * 23)
    P = np.sort(P)[::-1]
    n = P.shape[0]
    L = ['L{}'.format(i + 1) for i in range(n)]

    if pNone is None:
        print("Estimate p(None|x) as (1-p_1)*(1-p_2)*...*(1-p_n)")
        pNone = (1.0 - P).prod()

    PL = ['p({}|x)={}'.format(l, p) for l, p in zip(L, P)]
    print("Posteriors: {} (n={})".format(PL, n))
    print("p(None|x)={}".format(pNone))

    opt = F1Optimizer.maximize_expectation(P, pNone)
    best_prediction = ['None'] if opt[1] else []
    best_prediction += (L[:opt[0]])
    f1_max = opt[2]

    print("Prediction {} yields best E[F1] of {}\n".format(best_prediction, f1_max))


def save_plot(P, filename='expected_f1.png'):
    E_F1 = pd.DataFrame(F1Optimizer.get_expectations(P).T, columns=["/w None", "/wo None"])
    best_k, _, max_f1 = F1Optimizer.maximize_expectation(P)

    plt.style.use('ggplot')
    plt.figure()
    E_F1.plot()
    plt.title('Expected F1-Score for \n {}'.format("P = [{}]".format(",".join(map(str, P)))), fontsize=12)
    plt.xlabel('k')
    plt.xticks(np.arange(0, len(P) + 1, 1.0))
    plt.ylabel('E[F1(P,k)]')
    plt.plot([best_k], [max_f1], 'o', color='#000000', markersize=4)
    plt.annotate('max E[F1(P,k)] = E[F1(P,{})] = {:.5f}'.format(best_k, max_f1), xy=(best_k, max_f1),
                 xytext=(best_k, max_f1 * 0.8), arrowprops=dict(facecolor='black', shrink=0.05, width=1, headwidth=7),
                 horizontalalignment='center', verticalalignment='top')
    plt.gcf().savefig(filename)



def timeit(P):
    s = datetime.now()
    F1Optimizer.maximize_expectation(P)
    e = datetime.now()
    return (e-s).microseconds / 1E6


def benchmark(n=100, filename='runtimes.png'):
    results = pd.DataFrame(index=np.arange(1,n+1))
    results['runtimes'] = 0

    for i in range(1,n+1):
        runtimes = []
        for j in range(5):
            runtimes.append(timeit(np.sort(np.random.rand(i))[::-1]))
        results.iloc[i-1] = np.mean(runtimes)

    x = results.index
    y = results.runtimes
    results['quadratic fit'] = np.poly1d(np.polyfit(x, y, deg=2))(x)

    plt.style.use('ggplot')
    plt.figure()
    results.plot()
    plt.title('Expectation Maximization Runtimes', fontsize=12)
    plt.xlabel('n = |P|')
    plt.ylabel('time in seconds')
    plt.gcf().savefig(filename)


if __name__ == '__main__':
    print_best_prediction([0.3, 0.2])
    print_best_prediction([0.3, 0.2], 0.57)
    print_best_prediction([0.9, 0.6])
    print_best_prediction([0.5, 0.4, 0.3, 0.35, 0.33, 0.31, 0.29, 0.27, 0.25, 0.20, 0.15, 0.10])
    print_best_prediction([0.5, 0.4, 0.3, 0.35, 0.33, 0.31, 0.29, 0.27, 0.25, 0.20, 0.15, 0.10], 0.2)

    save_plot([0.45, 0.35, 0.31, 0.29, 0.27, 0.25, 0.22, 0.20, 0.17, 0.15, 0.10, 0.05, 0.02])
    benchmark()


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : utils
# @Time         : 2020-02-19 11:46
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 
import pandas as pd


def libsvm2df(file, max_index=16, batch_size=128):
    dic = dict(zip(range(max_index), [0] * max_index))
    fn = lambda r: {**dic, **eval("{0:" + r.replace(' ', ',') + "}")}
    df_iter = pd.read_csv(file, names='r', iterator=True)
    yield pd.DataFrame(df_iter.get_chunk(batch_size)['r'].map(fn).values.tolist())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : demo
# @Time         : 2019-09-10 15:17
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 

from yaml import safe_load

print(safe_load(open('./files.yml')))

from tensorflow.python.keras.preprocessing import text
from tensorflow.python.keras.preprocessing.text import hashing_trick


text.Tokenizer

from tensorflow.keras.preprocessing import text


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : backend
# @Time         : 2019-09-10 15:16
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


# import tensorflow as tf
#
# from tensorflow.python import keras
#
# utils = keras.utils
# activations = keras.activations
# applications = keras.applications
# backend = keras.backend
# datasets = keras.datasets
# engine = keras.engine
# layers = keras.layers
# preprocessing = keras.preprocessing
# wrappers = keras.wrappers
# callbacks = keras.callbacks
# constraints = keras.constraints
# initializers = keras.initializers
# metrics = keras.metrics
# models = keras.models
# losses = keras.losses
# optimizers = keras.optimizers
# regularizers = keras.regularizers



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : Python.
# @File         : inference
# @Time         : 2020-03-16 23:13
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 
import numpy as np
import onnxruntime as rt


class Inference(object):

    def __init__(self, path="./iris.onnx"):
        self.sess = rt.InferenceSession(path)
        self._describe()

    def run(self, X):
        _ = self.sess.get_inputs()[0]
        self.input_name = _.name
        # self.input_shape = tuple(_.shape)
        # assert X.shape == self.input_shape

        if not isinstance(X[0][0], np.float32):
            X = X.astype(np.float32)
        return self.sess.run(None, {self.input_name: X})  # 概率输出 or 类别输出

    def _describe(self):

        for attr_ in ['get_inputs', 'get_outputs']:
            _puts = self.sess.__getattribute__(attr_)()
            for i, _put in enumerate(_puts, 1):
                print({attr_.split('_')[-1][:-1].title() + str(i): (_put.type, _put.name, _put.shape)})



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Project      : DeepTricks.
# @File         : demo
# @Time         : 2020-02-12 20:17
# @Author       : yuanjie
# @Email        : yuanjie@xiaomi.com
# @Software     : PyCharm
# @Description  : 


import numpy as np
from sklearn.datasets import make_classification
from torch import nn
import torch.nn.functional as F

from skorch import NeuralNetClassifier


X, y = make_classification(1000, 20, n_informative=10, random_state=0)
X = X.astype(np.float32)
y = y.astype(np.int64)

class MyModule(nn.Module):
    def __init__(self, num_units=10, nonlin=F.relu):
        super(MyModule, self).__init__()

        self.dense0 = nn.Linear(20, num_units)
        self.nonlin = nonlin
        self.dropout = nn.Dropout(0.5)
        self.dense1 = nn.Linear(num_units, 10)
        self.output = nn.Linear(10, 2)

    def forward(self, X, **kwargs):
        X = self.nonlin(self.dense0(X))
        X = self.dropout(X)
        X = F.relu(self.dense1(X))
        X = F.softmax(self.output(X), dim=-1)
        return X


net = NeuralNetClassifier(
    MyModule,
    max_epochs=10,
    lr=0.1,
    # Shuffle training data on each epoch
    iterator_train__shuffle=True,
)

net.fit()
net.fit_loop()